// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information
use crate::{module::transform_body::Visitor, options, Errors, HdlAttr};
use proc_macro2::{Span, TokenStream};
use quote::{format_ident, quote_spanned, ToTokens, TokenStreamExt};
use syn::{
    parse::Nothing,
    parse_quote, parse_quote_spanned,
    punctuated::{Pair, Punctuated},
    spanned::Spanned,
    token::{Brace, Paren},
    Attribute, Expr, ExprArray, ExprCall, ExprGroup, ExprPath, ExprRepeat, ExprStruct, ExprTuple,
    FieldValue, Ident, Index, Member, Path, PathArguments, PathSegment, Token, TypePath,
};

options! {
    #[options = AggregateLiteralOptions]
    #[no_ident_fragment]
    pub(crate) enum AggregateLiteralOption {
        Struct(struct_),
        Enum(enum_),
    }
}

#[derive(Clone, Debug)]
pub(crate) struct StructOrEnumPath {
    pub(crate) ty: TypePath,
    pub(crate) variant: Option<(TypePath, Ident)>,
}

#[derive(Debug, Copy, Clone)]
pub(crate) struct SingleSegmentVariant {
    pub(crate) name: &'static str,
    pub(crate) make_type_path: fn(Span, &PathArguments) -> Path,
}

impl StructOrEnumPath {
    pub(crate) const SINGLE_SEGMENT_VARIANTS: &'static [SingleSegmentVariant] = {
        fn make_option_type_path(span: Span, arguments: &PathArguments) -> Path {
            let arguments = if arguments.is_none() {
                quote_spanned! {span=>
                    <_>
                }
            } else {
                arguments.to_token_stream()
            };
            parse_quote_spanned! {span=>
                ::fayalite::__std::option::Option #arguments
            }
        }
        fn make_result_type_path(span: Span, arguments: &PathArguments) -> Path {
            let arguments = if arguments.is_none() {
                quote_spanned! {span=>
                    <_, _>
                }
            } else {
                arguments.to_token_stream()
            };
            parse_quote_spanned! {span=>
                ::fayalite::__std::result::Result #arguments
            }
        }
        &[
            SingleSegmentVariant {
                name: "Some",
                make_type_path: make_option_type_path,
            },
            SingleSegmentVariant {
                name: "None",
                make_type_path: make_option_type_path,
            },
            SingleSegmentVariant {
                name: "Ok",
                make_type_path: make_result_type_path,
            },
            SingleSegmentVariant {
                name: "Err",
                make_type_path: make_result_type_path,
            },
        ]
    };
    pub(crate) fn new(
        errors: &mut Errors,
        path: TypePath,
        options: &AggregateLiteralOptions,
    ) -> Result<Self, ()> {
        let Path {
            leading_colon,
            segments,
        } = &path.path;
        let qself_position = path.qself.as_ref().map(|qself| qself.position).unwrap_or(0);
        let variant_name = if qself_position < segments.len() {
            Some(segments.last().unwrap().ident.clone())
        } else {
            None
        };
        let enum_type = 'guess_enum_type: {
            if options.enum_.is_some() {
                if let Some((struct_,)) = options.struct_ {
                    errors.error(
                        struct_,
                        "can't specify both #[hdl(enum)] and #[hdl(struct)]",
                    );
                }
                break 'guess_enum_type Some(None);
            }
            if options.struct_.is_some() {
                break 'guess_enum_type None;
            }
            if path.qself.is_none() && leading_colon.is_none() && segments.len() == 1 {
                let PathSegment { ident, arguments } = &segments[0];
                for &SingleSegmentVariant {
                    name,
                    make_type_path,
                } in Self::SINGLE_SEGMENT_VARIANTS
                {
                    if ident == name {
                        break 'guess_enum_type Some(Some(TypePath {
                            qself: None,
                            path: make_type_path(ident.span(), arguments),
                        }));
                    }
                }
            }
            if segments.len() == qself_position + 2
                && segments[qself_position + 1].arguments.is_none()
                && (path.qself.is_some()
                    || segments[qself_position].ident.to_string().as_bytes()[0]
                        .is_ascii_uppercase())
            {
                let mut ty = path.clone();
                ty.path.segments.pop();
                ty.path.segments.pop_punct();
                break 'guess_enum_type Some(Some(ty));
            }
            None
        };
        if let Some(enum_type) = enum_type {
            let ty = if let Some(enum_type) = enum_type {
                enum_type
            } else {
                if qself_position >= segments.len() {
                    errors.error(path, "#[hdl]: can't figure out enum's type");
                    return Err(());
                }
                let mut ty = path.clone();
                ty.path.segments.pop();
                ty.path.segments.pop_punct();
                ty
            };
            let Some(variant_name) = variant_name else {
                errors.error(path, "#[hdl]: can't figure out enum's variant name");
                return Err(());
            };
            Ok(Self {
                ty,
                variant: Some((path, variant_name)),
            })
        } else {
            Ok(Self {
                ty: path,
                variant: None,
            })
        }
    }
}

#[derive(Copy, Clone, Debug)]
pub(crate) enum BraceOrParen {
    Brace(Brace),
    Paren(Paren),
}

impl BraceOrParen {
    pub(crate) fn surround(self, tokens: &mut TokenStream, f: impl FnOnce(&mut TokenStream)) {
        match self {
            BraceOrParen::Brace(v) => v.surround(tokens, f),
            BraceOrParen::Paren(v) => v.surround(tokens, f),
        }
    }
}

#[derive(Debug, Clone)]
pub(crate) struct StructOrEnumLiteralField {
    pub(crate) attrs: Vec<Attribute>,
    pub(crate) member: Member,
    pub(crate) colon_token: Option<Token![:]>,
    pub(crate) expr: Expr,
}

#[derive(Debug, Clone)]
pub(crate) struct StructOrEnumLiteral {
    pub(crate) attrs: Vec<Attribute>,
    pub(crate) path: TypePath,
    pub(crate) brace_or_paren: BraceOrParen,
    pub(crate) fields: Punctuated<StructOrEnumLiteralField, Token![,]>,
    pub(crate) dot2_token: Option<Token![..]>,
    pub(crate) rest: Option<Box<Expr>>,
}

impl StructOrEnumLiteral {
    pub(crate) fn map_field_exprs(self, mut f: impl FnMut(Expr) -> Expr) -> Self {
        self.map_fields(|mut field| {
            field.expr = f(field.expr);
            field
        })
    }
    pub(crate) fn map_fields(
        self,
        mut f: impl FnMut(StructOrEnumLiteralField) -> StructOrEnumLiteralField,
    ) -> Self {
        let Self {
            attrs,
            path,
            brace_or_paren,
            fields,
            dot2_token,
            rest,
        } = self;
        let fields = Punctuated::from_iter(fields.into_pairs().map(|p| {
            let (field, comma) = p.into_tuple();
            Pair::new(f(field), comma)
        }));
        Self {
            attrs,
            path,
            brace_or_paren,
            fields,
            dot2_token,
            rest,
        }
    }
}

impl From<ExprStruct> for StructOrEnumLiteral {
    fn from(value: ExprStruct) -> Self {
        let ExprStruct {
            attrs,
            qself,
            path,
            brace_token,
            fields,
            dot2_token,
            rest,
        } = value;
        Self {
            attrs,
            path: TypePath { qself, path },
            brace_or_paren: BraceOrParen::Brace(brace_token),
            fields: Punctuated::from_iter(fields.into_pairs().map(|v| {
                let (
                    FieldValue {
                        attrs,
                        member,
                        colon_token,
                        expr,
                    },
                    comma,
                ) = v.into_tuple();
                Pair::new(
                    StructOrEnumLiteralField {
                        attrs,
                        member,
                        colon_token,
                        expr,
                    },
                    comma,
                )
            })),
            dot2_token,
            rest,
        }
    }
}

fn expr_to_member(expr: &Expr) -> Option<Member> {
    syn::parse2(expr.to_token_stream()).ok()
}

impl ToTokens for StructOrEnumLiteral {
    fn to_tokens(&self, tokens: &mut TokenStream) {
        let Self {
            attrs,
            path,
            brace_or_paren,
            fields,
            dot2_token,
            rest,
        } = self;
        tokens.append_all(attrs);
        path.to_tokens(tokens);
        brace_or_paren.surround(tokens, |tokens| {
            match brace_or_paren {
                BraceOrParen::Brace(_) => {
                    for (
                        StructOrEnumLiteralField {
                            attrs,
                            member,
                            mut colon_token,
                            expr,
                        },
                        comma,
                    ) in fields.pairs().map(|v| v.into_tuple())
                    {
                        tokens.append_all(attrs);
                        if Some(member) != expr_to_member(expr).as_ref() {
                            colon_token = Some(<Token![:]>::default());
                        }
                        member.to_tokens(tokens);
                        colon_token.to_tokens(tokens);
                        expr.to_tokens(tokens);
                        comma.to_tokens(tokens);
                    }
                }
                BraceOrParen::Paren(_) => {
                    for (
                        StructOrEnumLiteralField {
                            attrs,
                            member: _,
                            colon_token: _,
                            expr,
                        },
                        comma,
                    ) in fields.pairs().map(|v| v.into_tuple())
                    {
                        tokens.append_all(attrs);
                        expr.to_tokens(tokens);
                        comma.to_tokens(tokens);
                    }
                }
            }
            if let Some(rest) = rest {
                dot2_token.unwrap_or_default().to_tokens(tokens);
                rest.to_tokens(tokens);
            }
        });
    }
}

impl Visitor {
    pub(crate) fn process_hdl_array(
        &mut self,
        hdl_attr: HdlAttr<Nothing>,
        mut expr_array: ExprArray,
    ) -> Expr {
        self.require_normal_module(hdl_attr);
        for elem in &mut expr_array.elems {
            *elem = parse_quote_spanned! {elem.span()=>
                ::fayalite::expr::ToExpr::to_expr(&(#elem))
            };
        }
        parse_quote! {::fayalite::expr::ToExpr::to_expr(&#expr_array)}
    }
    pub(crate) fn process_hdl_repeat(
        &mut self,
        hdl_attr: HdlAttr<Nothing>,
        mut expr_repeat: ExprRepeat,
    ) -> Expr {
        self.require_normal_module(hdl_attr);
        let repeated_value = &expr_repeat.expr;
        *expr_repeat.expr = parse_quote_spanned! {repeated_value.span()=>
            ::fayalite::expr::ToExpr::to_expr(&(#repeated_value))
        };
        parse_quote! {::fayalite::expr::ToExpr::to_expr(&#expr_repeat)}
    }
    pub(crate) fn process_struct_enum(
        &mut self,
        hdl_attr: HdlAttr<AggregateLiteralOptions>,
        mut literal: StructOrEnumLiteral,
    ) -> Expr {
        let span = hdl_attr.hdl.span;
        if let Some(rest) = literal.rest.take() {
            self.errors
                .error(rest, "#[hdl] struct functional update syntax not supported");
        }
        let mut next_var = 0usize;
        let mut new_var = || -> Ident {
            let retval = format_ident!("__v{}", next_var, span = span);
            next_var += 1;
            retval
        };
        let infallible_var = new_var();
        let retval_var = new_var();
        let mut lets = vec![];
        let mut build_steps = vec![];
        let literal = literal.map_field_exprs(|expr| {
            let field_var = new_var();
            lets.push(quote_spanned! {span=>
                let #field_var = ::fayalite::expr::ToExpr::to_expr(&#expr);
            });
            parse_quote! { #field_var }
        });
        let Ok(StructOrEnumPath { ty, variant }) =
            StructOrEnumPath::new(&mut self.errors, literal.path.clone(), &hdl_attr.body)
        else {
            return parse_quote_spanned! {span=>
                {}
            };
        };
        for StructOrEnumLiteralField {
            attrs: _,
            member,
            colon_token: _,
            expr,
        } in literal.fields.iter()
        {
            let field_fn = format_ident!("field_{}", member);
            build_steps.push(quote_spanned! {span=>
                let #retval_var = #retval_var.#field_fn(#expr);
            });
        }
        let check_literal = literal.map_field_exprs(|expr| {
            parse_quote_spanned! {span=>
                ::fayalite::expr::value_from_expr_type(#expr, #infallible_var)
            }
        });
        let make_expr_fn = if let Some((_variant_path, variant_ident)) = &variant {
            let variant_fn = format_ident!("variant_{}", variant_ident);
            build_steps.push(quote_spanned! {span=>
                let #retval_var = #retval_var.#variant_fn();
            });
            quote_spanned! {span=>
                ::fayalite::expr::make_enum_expr
            }
        } else {
            build_steps.push(quote_spanned! {span=>
                let #retval_var = #retval_var.build();
            });
            quote_spanned! {span=>
                ::fayalite::expr::make_bundle_expr
            }
        };
        let variant_or_type =
            variant.map_or_else(|| ty.clone(), |(variant_path, _variant_ident)| variant_path);
        parse_quote_spanned! {span=>
            {
                #(#lets)*
                #make_expr_fn::<#ty>(|#infallible_var| {
                    let #retval_var = #check_literal;
                    #[allow(unreachable_code)]
                    match #retval_var {
                        #variant_or_type { .. } => #retval_var,
                        #[allow(unreachable_patterns)]
                        _ => match #infallible_var {},
                    }
                }, |#retval_var| {
                    #(#build_steps)*
                    #retval_var
                })
            }
        }
    }
    pub(crate) fn process_hdl_struct(
        &mut self,
        hdl_attr: HdlAttr<AggregateLiteralOptions>,
        expr_struct: ExprStruct,
    ) -> Expr {
        self.require_normal_module(&hdl_attr);
        self.process_struct_enum(hdl_attr, expr_struct.into())
    }
    pub(crate) fn process_hdl_tuple(
        &mut self,
        hdl_attr: HdlAttr<Nothing>,
        expr_tuple: ExprTuple,
    ) -> Expr {
        self.require_normal_module(hdl_attr);
        parse_quote_spanned! {expr_tuple.span()=>
            ::fayalite::expr::ToExpr::to_expr(&#expr_tuple)
        }
    }
    pub(crate) fn process_hdl_path(
        &mut self,
        hdl_attr: HdlAttr<Nothing>,
        expr_path: ExprPath,
    ) -> Expr {
        self.require_normal_module(hdl_attr);
        parse_quote_spanned! {expr_path.span()=>
            ::fayalite::expr::ToExpr::to_expr(&#expr_path)
        }
    }
    pub(crate) fn process_hdl_call(
        &mut self,
        hdl_attr: HdlAttr<AggregateLiteralOptions>,
        expr_call: ExprCall,
    ) -> Expr {
        self.require_normal_module(&hdl_attr);
        let ExprCall {
            attrs: mut literal_attrs,
            func,
            paren_token,
            args,
        } = expr_call;
        let mut path_expr = *func;
        let path = loop {
            break match path_expr {
                Expr::Group(ExprGroup {
                    attrs,
                    group_token: _,
                    expr,
                }) => {
                    literal_attrs.extend(attrs);
                    path_expr = *expr;
                    continue;
                }
                Expr::Path(ExprPath { attrs, qself, path }) => {
                    literal_attrs.extend(attrs);
                    TypePath { qself, path }
                }
                _ => {
                    self.errors.error(&path_expr, "missing tuple struct's name");
                    return parse_quote_spanned! {path_expr.span()=>
                        {}
                    };
                }
            };
        };
        let fields = Punctuated::from_iter(args.into_pairs().enumerate().map(|(index, p)| {
            let (expr, comma) = p.into_tuple();
            let mut index = Index::from(index);
            index.span = hdl_attr.hdl.span;
            Pair::new(
                StructOrEnumLiteralField {
                    attrs: vec![],
                    member: Member::Unnamed(index),
                    colon_token: None,
                    expr,
                },
                comma,
            )
        }));
        self.process_struct_enum(
            hdl_attr,
            StructOrEnumLiteral {
                attrs: literal_attrs,
                path,
                brace_or_paren: BraceOrParen::Paren(paren_token),
                fields,
                dot2_token: None,
                rest: None,
            },
        )
    }
}
